<?php
/************************************************************************
 * Mailer v0.2.1-FINAL                                Start: 03/08/2011 *
 * ===================                          Last change: 03/08/2011 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : http-functions.php                               *
 * -------------------------------------------------------------------- *
 * Short description : HTTP-related functions                           *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : HTTP-relevante Funktionen                        *
 * -------------------------------------------------------------------- *
 * $Revision::                                                        $ *
 * $Date::                                                            $ *
 * $Tag:: 0.2.1-FINAL                                                 $ *
 * $Author::                                                          $ *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2009 by Roland Haeder                           *
 * Copyright (c) 2009 - 2011 by Mailer Developer Team                   *
 * For more information visit: http://mxchange.org                      *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	die();
} // END - if

// Sends out all headers required for HTTP/1.1 reply
function sendHttpHeaders () {
	// Used later
	$now = gmdate('D, d M Y H:i:s') . ' GMT';

	// Send HTTP header
	sendHeader('HTTP/1.1 ' . getHttpStatus());

	// General headers for no caching
	sendHeader('Expires: ' . $now); // RFC2616 - Section 14.21
	sendHeader('Last-Modified: ' . $now);
	sendHeader('Cache-Control: no-store, no-cache, must-revalidate, pre-check=0, post-check=0, max-age=0'); // HTTP/1.1
	sendHeader('Pragma: no-cache'); // HTTP/1.0
	sendHeader('Connection: Close');
	sendHeader('Content-Type: ' . getContentType() . '; charset=UTF-8');
	sendHeader('Content-Language: ' . getLanguage());
}

// Generates the full GET URL from given base URL and data array
function generateGetUrlFromBaseUrlData ($baseUrl, $requestData = array()) {
	// Init URL
	$getUrl = $baseUrl;

	// Add data
	$body = http_build_query($requestData, '', '&');

	// There should be data, else we don't need to extend $baseUrl with $body
	if (!empty($body)) {
		// Do we have a question-mark in the script?
		if (!isInString('?', $baseUrl)) {
			// No, so first char must be question mark
			$body = '?' . $body;
		} else {
			// Ok, add &
			$body = '&' . $body;
		}

		// Add script data
		$getUrl .= $body;

		// Remove trailed & to make it more conform
		if (substr($getUrl, -1, 1) == '&') {
			$getUrl = substr($getUrl, 0, -1);
		} // END - if
	} // END - if

	// Return it
	return $getUrl;
}

// Removes http[s]://<hostname> from given url
function removeHttpHostNameFromUrl ($url) {
	// Remove http[s]://
	$remove = explode(':', $url);
	$remove = explode('/', substr($remove[1], 3));

	// Remove the first element (should be the hostname)
	unset($remove[0]);

	// implode() back all other elements and prepend a slash
	$url = '/' . implode('/', $remove);

	// Return prepared URL
	return $url;
}

// Send a HEAD request
function sendHeadRequest ($baseUrl, $requestData = array()) {
	// Generate full GET URL
	$getUrl = generateGetUrlFromBaseUrlData($baseUrl, $requestData);

	// Extract hostname and port from script
	$host = extractHostnameFromUrl($baseUrl);

	// Remove http[s]://<hostname> from url
	$getUrl = removeHttpHostNameFromUrl($getUrl);

	// Generate HEAD request header
	$request  = 'HEAD ' . trim($getUrl) . ' HTTP/1.1' . getConfig('HTTP_EOL');
	$request .= 'Host: ' . $host . getConfig('HTTP_EOL');
	$request .= 'Referer: ' . getUrl() . '/admin.php' . getConfig('HTTP_EOL');
	if (isConfigEntrySet('FULL_VERSION')) {
		$request .= 'User-Agent: ' . getTitle() . '/' . getFullVersion() . getConfig('HTTP_EOL');
	} else {
		$request .= 'User-Agent: ' . getTitle() . '/' . getConfig('VERSION') . getConfig('HTTP_EOL');
	}
	$request .= 'Accept: image/png,image/*;q=0.8,text/plain,text/html,*/*;q=0.5' . getConfig('HTTP_EOL');
	$request .= 'Accept-Charset: UTF-8,*' . getConfig('HTTP_EOL');
	$request .= 'Cache-Control: no-store, no-cache, must-revalidate, post-check=0, pre-check=0' . getConfig('HTTP_EOL');
	$request .= 'Connection: close' . getConfig('HTTP_EOL');
	$request .= getConfig('HTTP_EOL');

	// Send the raw request
	$response = sendRawRequest($host, $request);

	// Return the result to the caller function
	return $response;
}

// Send a GET request
function sendGetRequest ($baseUrl, $requestData = array(), $removeHeader = false) {
	// Generate full GET URL
	$getUrl = generateGetUrlFromBaseUrlData($baseUrl, $requestData);

	// Extract hostname and port from script
	$host = extractHostnameFromUrl($baseUrl);

	// Remove http[s]://<hostname> from url
	$getUrl = removeHttpHostNameFromUrl($getUrl);

	// Generate GET request header
	$request  = 'GET ' . trim($getUrl) . ' HTTP/1.1' . getConfig('HTTP_EOL');
	$request .= 'Host: ' . $host . getConfig('HTTP_EOL');
	$request .= 'Referer: ' . getUrl() . '/admin.php' . getConfig('HTTP_EOL');
	if (isConfigEntrySet('FULL_VERSION')) {
		$request .= 'User-Agent: ' . getTitle() . '/' . getFullVersion() . getConfig('HTTP_EOL');
	} else {
		$request .= 'User-Agent: ' . getTitle() . '/' . getConfig('VERSION') . getConfig('HTTP_EOL');
	}
	$request .= 'Accept: image/png,image/*;q=0.8,text/plain,text/html,*/*;q=0.5' . getConfig('HTTP_EOL');
	$request .= 'Accept-Charset: UTF-8,*' . getConfig('HTTP_EOL');
	$request .= 'Cache-Control: no-store, no-cache, must-revalidate, post-check=0, pre-check=0' . getConfig('HTTP_EOL');
	$request .= 'Connection: close' . getConfig('HTTP_EOL');
	$request .= getConfig('HTTP_EOL');

	// Send the raw request
	$response = sendRawRequest($host, $request);

	// Should we remove header lines?
	if ($removeHeader === true) {
		// Okay, remove them
		$response = removeHttpHeaderFromResponse($response);
	} // END - if

	// Return the result to the caller function
	return $response;
}

// Send a POST request
function sendPostRequest ($baseUrl, $requestData, $removeHeader = false) {
	// Extract host name from script
	$host = extractHostnameFromUrl($baseUrl);

	// Construct request body
	$body = http_build_query($requestData, '', '&');

	// Remove http(s)://$host from base URL
	$baseUrl = removeHttpHostNameFromUrl($baseUrl);

	// Generate POST request header
	$request  = 'POST ' . trim($baseUrl) . ' HTTP/1.0' . getConfig('HTTP_EOL');
	$request .= 'Host: ' . $host . getConfig('HTTP_EOL');
	$request .= 'Referer: ' . getUrl() . '/admin.php' . getConfig('HTTP_EOL');
	if (isConfigEntrySet('FULL_VERSION')) {
		$request .= 'User-Agent: ' . getTitle() . '/' . getFullVersion() . getConfig('HTTP_EOL');
	} else {
		$request .= 'User-Agent: ' . getTitle() . '/' . getConfig('VERSION') . getConfig('HTTP_EOL');
	}
	$request .= 'Accept: text/plain;q=0.8' . getConfig('HTTP_EOL');
	$request .= 'Accept-Charset: UTF-8,*' . getConfig('HTTP_EOL');
	$request .= 'Cache-Control: no-store, no-cache, must-revalidate, post-check=0, pre-check=0' . getConfig('HTTP_EOL');
	$request .= 'Content-Type: application/x-www-form-urlencoded' . getConfig('HTTP_EOL');
	$request .= 'Content-Length: ' . strlen($body) . getConfig('HTTP_EOL');
	$request .= 'Connection: close' . getConfig('HTTP_EOL');
	$request .= getConfig('HTTP_EOL');

	// Add body
	$request .= $body;

	// Send the raw request
	$response = sendRawRequest($host, $request);

	// Should we remove header lines?
	if ($removeHeader === true) {
		// Okay, remove them
		$response = removeHttpHeaderFromResponse($response);
	} // END - if

	// Return the result to the caller function
	return $response;
}

// Sends a raw request to another host
function sendRawRequest ($host, $request) {
	//* DEBUG: */ die('host='.$host.',request=<pre>'.$request.'</pre>');
	// Init errno and errdesc with 'all fine' values
	$errno = '0';
	$errdesc = '';

	// Default port is 80
	$port = 80;

	// Initialize array
	$response = array('', '', '');

	// Default is not to use proxy
	$useProxy = false;

	// Default is non-broken HTTP server implementation
	$GLOBALS['is_http_server_broken'] = false;

	// Are proxy settins set?
	if (isProxyUsed()) {
		// Then use it
		$useProxy = true;
	} // END - if

	// Load include
	loadIncludeOnce('inc/classes/resolver.class.php');

	// Extract port part from host
	$portArray = explode(':', $host);
	if (count($portArray) == 2) {
		// Extract host and port
		$host = $portArray[0];
		$port = $portArray[1];
	} elseif (count($portArray) > 2) {
		// This should not happen!
		debug_report_bug(__FUNCTION__, __LINE__, 'Invalid ' . $host . '. Please report this to the Mailer-Project team.');
	}

	// Get resolver instance
	$resolver = new HostnameResolver();

	// Open connection
	if ($useProxy === true) {
		// Resolve hostname into IP address
		$ip = $resolver->resolveHostname(compileRawCode(getProxyHost()));

		// Connect to host through proxy connection
		$fp = fsockopen($ip, bigintval(getProxyPort()), $errno, $errdesc, 30);
	} else {
		// Resolve hostname into IP address
		$ip = $resolver->resolveHostname($host);

		// Connect to host directly
		$fp = fsockopen($ip, $port, $errno, $errdesc, 30);
	}

	// Is there a link?
	if (!is_resource($fp)) {
		// Failed!
		logDebugMessage(__FUNCTION__, __LINE__, $errdesc . ' (' . $errno . ')');
		return $response;
	} elseif ((!stream_set_blocking($fp, 0)) || (!stream_set_timeout($fp, 1))) {
		// Cannot set non-blocking mode or timeout
		logDebugMessage(__FUNCTION__, __LINE__, socket_strerror(socket_last_error()));
		return $response;
	}

	// Do we use proxy?
	if ($useProxy === true) {
		// Setup proxy tunnel
		$response = setupProxyTunnel($host, $port, $fp);

		// If the response is invalid, abort
		if ((count($response) == 3) && (empty($response[0])) && (empty($response[1])) && (empty($response[2]))) {
			// Invalid response!
			logDebugMessage(__FUNCTION__, __LINE__, 'Proxy tunnel not working?');
			return $response;
		} // END - if
	} // END - if

	// Write request
	fwrite($fp, $request);

	// Start counting
	$start = microtime(true);

	// Read response
	while (!feof($fp)) {
		// Get info from stream
		$info = stream_get_meta_data($fp);

		// Is it timed out? 15 seconds is a really patient...
		if (($info['timed_out'] == true) || (microtime(true) - $start) > 15) {
			// Timeout
			logDebugMessage(__FUNCTION__, __LINE__, 'Timed out to get data from host ' . $host);

			// Abort here
			break;
		} // END - if

		// Get line from stream
		$line = fgets($fp, 128);

		// Ignore empty lines because of non-blocking mode
		if (empty($line)) {
			// uslepp a little to avoid 100% CPU load
			usleep(10);

			// Skip this
			continue;
		} // END - if

		// Check for broken HTTP implementations
		if (substr(strtolower($line), 0, 7) == 'server:') {
			// Anomic (see http://anomic.de, http://yacy.net) is currently broken
			$GLOBALS['is_http_server_broken'] = (count(getArrayKeysFromSubStrArray(strtolower($line), array('anomichttpd'))) > 0);
		} // END - if

		// Add it to response
		//* DEBUG: */ print 'line='.$line.'<br />';
		$response[] = $line;
	} // END - while

	// Close socket
	fclose($fp);

	// Time request if debug-mode is enabled
	if (isDebugModeEnabled()) {
		// Add debug message...
		logDebugMessage(__FUNCTION__, __LINE__, 'Request took ' . (microtime(true) - $start) . ' seconds and returned ' . count($response) . ' line(s).');
	} // END - if

	// Skip first empty lines
	$resp = $response;
	foreach ($resp as $idx => $line) {
		// Trim space away
		$line = trim($line);

		// Is this line empty?
		if (empty($line)) {
			// Then remove it
			array_shift($response);
		} else {
			// Abort on first non-empty line
			break;
		}
	} // END - foreach

	//* DEBUG: */ debugOutput('<strong>Request:</strong><pre>'.print_r($request, true).'</pre>');
	//* DEBUG: */ debugOutput('<strong>Response:</strong><pre>'.print_r($response, true).'</pre>');

	// Proxy agent found or something went wrong?
	if (!isset($response[0])) {
		// No response, maybe timeout
		$response = array('', '', '');
		logDebugMessage(__FUNCTION__, __LINE__, 'Invalid empty response array, maybe timed out?');
	} elseif ((substr(strtolower($response[0]), 0, 11) == 'proxy-agent') && ($useProxy === true)) {
		// Proxy header detected, so remove two lines
		array_shift($response);
		array_shift($response);
	} // END - if

	// Was the request successfull?
	if ((!isInStringIgnoreCase('200 OK', $response[0])) || (empty($response[0]))) {
		// Not found / access forbidden
		logDebugMessage(__FUNCTION__, __LINE__, 'Unexpected status code ' . $response[0] . ' detected. "200 OK" was expected.');
		$response = array('', '', '');
	} else {
		// Check array for chuncked encoding
		$response = unchunkHttpResponse($response);
	} // END - if

	// Return response
	return $response;
}

// Sets up a proxy tunnel for given hostname and through resource
function setupProxyTunnel ($host, $port, $resource) {
	// Initialize array
	$response = array('', '', '');

	// Generate CONNECT request header
	$proxyTunnel  = 'CONNECT ' . $host . ':' . $port . ' HTTP/1.0' . getConfig('HTTP_EOL');
	$proxyTunnel .= 'Host: ' . $host . getConfig('HTTP_EOL');

	// Use login data to proxy? (username at least!)
	if (getProxyUsername() != '') {
		// Add it as well
		$encodedAuth = base64_encode(compileRawCode(getProxyUsername()) . ':' . compileRawCode(getProxyPassword()));
		$proxyTunnel .= 'Proxy-Authorization: Basic ' . $encodedAuth . getConfig('HTTP_EOL');
	} // END - if

	// Add last new-line
	$proxyTunnel .= getConfig('HTTP_EOL');
	//* DEBUG: */ debugOutput('<strong>proxyTunnel=</strong><pre>' . $proxyTunnel.'</pre>');

	// Write request
	fwrite($fp, $proxyTunnel);

	// Got response?
	if (feof($fp)) {
		// No response received
		return $response;
	} // END - if

	// Read the first line
	$resp = trim(fgets($fp, 10240));
	$respArray = explode(' ', $resp);
	if ((strtolower($respArray[0]) !== 'http/1.0') || ($respArray[1] != '200')) {
		// Invalid response!
		return $response;
	} // END - if

	// All fine!
	return $respArray;
}

// Check array for chuncked encoding
function unchunkHttpResponse ($response) {
	// Default is not chunked
	$isChunked = false;

	// Check if we have chunks
	foreach ($response as $line) {
		// Make lower-case and trim it
		$line = trim($line);

		// Entry found?
		if ((isInStringIgnoreCase('transfer-encoding', $line)) && (isInStringIgnoreCase('chunked', $line))) {
			// Found!
			$isChunked = true;
			break;
		} // END - if
	} // END - foreach

	// Is it chunked?
	if ($isChunked === true) {
		// Good, we still have the HTTP headers in there, so we need to get rid
		// of them temporarly
		//* DEBUG: */ die('<pre>'.htmlentities(print_r(removeHttpHeaderFromResponse($response), true)).'</pre>');
		$tempResponse = http_chunked_decode(implode('', removeHttpHeaderFromResponse($response)));

		// We got a string back from http_chunked_decode(), so we need to convert it back to an array
		//* DEBUG: */ die('tempResponse['.strlen($tempResponse).']=<pre>'.replaceReturnNewLine(htmlentities($tempResponse)).'</pre>');

		// Re-add the headers
		$response = merge_array($GLOBALS['http_headers'], stringToArray("\n", $tempResponse));
	} // END - if

	// Return the unchunked array
	return $response;
}

// Removes HTTP header lines from a response array (e.g. output from send<Get|Post>Request() )
function removeHttpHeaderFromResponse ($response) {
	// Save headers for later usage
	$GLOBALS['http_headers'] = array();

	// The first array element has to contain HTTP
	if ((isset($response[0])) && (substr(strtoupper($response[0]), 0, 5) == 'HTTP/')) {
		// Okay, we have headers, now remove them with a second array
		$response2 = $response;
		foreach ($response as $line) {
			// Remove line
			array_shift($response2);

			// Add full line to temporary global array
			$GLOBALS['http_headers'][] = $line;

			// Trim it for testing
			$lineTest = trim($line);

			// Is this line empty?
			if (empty($lineTest)) {
				// Then stop here
				break;
			} // END - if
		} // END - foreach

		// Write back the array
		$response = $response2;
	} // END - if

	// Return the modified response array
	return $response;
}

// Returns the flag if a broken HTTP server implementation was detected
function isBrokenHttpServerImplentation () {
	// Determine it
	$isBroken = ((isset($GLOBALS['is_http_server_broken'])) && ($GLOBALS['is_http_server_broken'] === true));

	// ... and return it
	return $isBroken;
}

//-----------------------------------------------------------------------------
// Automatically re-created functions, all taken from user comments on www.php.net
//-----------------------------------------------------------------------------

if (!function_exists('http_build_query')) {
	// Taken from documentation on www.php.net, credits to Marco K. (Germany) and some light mods by R.Haeder
	function http_build_query($requestData, $prefix = '', $sep = '', $key = '') {
		$ret = array();
		foreach ((array) $requestData as $k => $v) {
			if (is_int($k) && $prefix != null) {
				$k = urlencode($prefix . $k);
			} // END - if

			if ((!empty($key)) || ($key === 0)) {
				$k = $key . '[' . urlencode($k) . ']';
			} // END - if

			if (is_array($v) || is_object($v)) {
				array_push($ret, http_build_query($v, '', $sep, $k));
			} else {
				array_push($ret, $k . '=' . urlencode($v));
			}
		} // END - foreach

		if (empty($sep)) {
			$sep = ini_get('arg_separator.output');
		} // END - if

		return implode($sep, $ret);
	}
} // END - if

if (!function_exists('http_chunked_decode')) {
	/**
	 * dechunk an HTTP 'transfer-encoding: chunked' message.
	 *
	 * @param	$chunk		The encoded message
	 * @return	$dechunk	The decoded message. If $chunk wasn't encoded properly debug_report_bug() is being called
	 * @author	Marques Johansson (initial author)
	 * @author	Roland Haeder (heavy modifications and simplification)
	 * @link	http://php.net/manual/en/function.http-chunked-decode.php#89786
	 */
	function http_chunked_decode ($chunk) {
		// Detect multi-byte encoding
		$mbPrefix = detectMultiBytePrefix($chunk);
		//* DEBUG: */ print 'mbPrefix=' . $mbPrefix . '<br />';

		// Init some variables
		$offset = 0;
		$len = call_user_func_array($mbPrefix . 'strlen', array(($chunk)));
		$dechunk = '';

		// Walk through all chunks
		while ($offset < $len) {
			// Where does the \r\n begin?
			$lineEndAt = call_user_func_array($mbPrefix . 'strpos', array($chunk, getConfig('HTTP_EOL'), $offset));

			/* DEBUG: *
			print 'lineEndAt[<em>'.__LINE__.'</em>]='.$lineEndAt.'<br />
offset[<em>'.__LINE__.'</em>]='.$offset.'<br />
len='.$len.'<br />
next[offset,10]=<pre>'.replaceReturnNewLine(htmlentities(call_user_func_array($mbPrefix . 'substr', array($chunk, $offset, 10)))).'</pre>';
			/* DEBUG: */

			// Get next hex-coded chunk length
			$chunkLenHex = call_user_func_array($mbPrefix . 'substr', array($chunk, $offset, ($lineEndAt - $offset)));

			/* DEBUG: *
			print 'chunkLenHex[<em>'.__LINE__.'</em>]='.replaceReturnNewLine(htmlentities($chunkLenHex)).'<br />
';
			/* DEBUG: */

			// Validation if it is hexadecimal
			if (!isHexadecimal($chunkLenHex)) {
				// Please help debugging this
				//* DEBUG: */ die('ABORT:chunkLenHex=<pre>'.replaceReturnNewLine(htmlentities($chunkLenHex)).'</pre>');
				debug_report_bug(__FUNCTION__, __LINE__, 'Value ' . $chunkLenHex . ' is no valid hexa-decimal string.');

				// This won't be reached
				return $chunk;
			} // END - if

			// Position of next chunk is right after \r\n
			$offset   = $offset + strlen($chunkLenHex) + strlen(getConfig('HTTP_EOL'));
			$chunkLen = hexdec(rtrim($chunkLenHex, getConfig('HTTP_EOL')));

			/* DEBUG: *
			print 'chunkLen='.$chunkLen.'<br />
offset[<em>'.__LINE__.'</em>]='.$offset.'<br />';
			/* DEBUG: */

			// Moved out for debugging
			$next  = call_user_func_array($mbPrefix . 'substr', array($chunk, $offset, $chunkLen));
			//* DEBUG: */ print 'next=<pre>'.replaceReturnNewLine(htmlentities($next)).'</pre>';

			/*
			 * Hack for e.g. YaCy HTTPDaemon (Anomic Server), this HTTP server
			 * is currently (revision 7567 and maybe earlier) broken and does
			 * not include the \r\n characters when it sents a "chunked" HTTP
			 * message.
			 */
			$count = 0;
			if (isBrokenHttpServerImplentation()) {
				// Count occurrences of \r\n
				$count = call_user_func_array($mbPrefix . 'substr_count', array($next, getConfig('HTTP_EOL')));
			} // END - if

			/*
			 * Correct chunk length because some broken HTTP server
			 * implementation subtract occurrences of \r\n in their chunk
			 * lengths.
			 */
			$chunkLen = hexdec(rtrim($chunkLenHex, getConfig('HTTP_EOL'))) - ($count * strlen(getConfig('HTTP_EOL')));

			// Add next chunk to $dechunk
			$dechunk .= call_user_func_array($mbPrefix . 'substr', array($chunk, $offset, $chunkLen));

			/* DEBUG: *
			print('offset[<em>'.__LINE__.'</em>]='.$offset.'<br />
lineEndAt[<em>'.__LINE__.'</em>]='.$lineEndAt.'<br />
len='.$len.'<br />
count='.$count.'<br />
chunkLen='.$chunkLen.'<br />
chunkLenHex='.$chunkLenHex.'<br />
dechunk=<pre>'.replaceReturnNewLine(htmlentities($dechunk)).'</pre>
chunk=<pre>'.replaceReturnNewLine(htmlentities($chunk)).'</pre>');
			/* DEBUG: */

			// Is $offset + $chunkLen larger than or equal $len?
			if (($offset + $chunkLen) >= $len) {
				// Then stop processing here
				break;
			} // END - if

			// Calculate offset of next chunk
			$offset = call_user_func_array($mbPrefix . 'strpos', array($chunk, getConfig('HTTP_EOL'), $offset + $chunkLen)) + 2;

			/* DEBUG: *
			print('offset[<em>'.__LINE__.'</em>]='.$offset.'<br />
next[100]=<pre>'.replaceReturnNewLine(htmlentities(call_user_func_array($mbPrefix . 'substr', array($chunk, $offset, 100)))).'</pre>
---:---:---:---:---:---:---:---:---<br />
');
			/* DEBUG: */
		} // END - while

		// Return de-chunked string
		return $dechunk;
	}
} // END - if

// [EOF]
?>
